#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdint.h>
#include <math.h>
#include <immintrin.h>

#define TIME2USEC(t) ((long)(t).tv_sec * 1000000L + (t).tv_usec)
#define TIME_DIFF(t1, t2) (TIME2USEC (t2) - TIME2USEC (t1))

#define ITERATIONS 100
#define N          (128 * 1000)

static const float epsilon = 2.0e-4;

static const float K = .025;

static float v1[N] __attribute__((aligned(64)));
static float v2[N] __attribute__((aligned(64)));
static float v3[N] __attribute__((aligned(64)));
static float ref[N] __attribute__((aligned(64)));

static void init_vectors (void)
{
  for (int i = 0; i < N; i++) {
    v1[i] = (float)i;
    v2[i] = 1.0 + (i % 2);
    v3[i] = 0.0;
    ref[i] = 0.0;
  }
}

static inline float f (float x1, float x2, float x3)
{
  float tmp = sqrt (K * x1 + x2);
  if (tmp < 10.0)
    return 1.0f / (10.0 + tmp) + x3;
  else
    return 1.0f / tmp + x3;
}

void compute (void)
{
#if AVX
    __m256 valK = _mm256_set1_ps (K);
    __m256 un = _mm256_set1_ps (1.0);
    __m256 dix = _mm256_set1_ps (10.0);

    for (int i = 0; i < N; i += 8) {
      __m256 val1 = _mm256_load_ps (&v1[i]);
      __m256 val2 = _mm256_load_ps (&v2[i]);
      __m256 val3 = _mm256_load_ps (&v3[i]);

      __m256 tmp = _mm256_fmadd_ps (valK, val1, val2);
      tmp = _mm256_sqrt_ps (tmp);

      // tmp < 10 ?
      __m256 mask = _mm256_cmp_ps(tmp, dix, _CMP_LT_OS);
      tmp = _mm256_add_ps (tmp, _mm256_and_ps (dix, mask));

      tmp = _mm256_div_ps (un, tmp);
      tmp = _mm256_add_ps (tmp, val3);
      _mm256_store_ps (&v3[i], tmp);
    }
#else
    for (int i = 0; i < N; i++)
      v3[i] = f(v1[i], v2[i], v3[i]);
#endif
}

static void check_result (void);
static void show_first_results (void);

int main(int argc, char **argv)
{
  struct timeval t1, t2;
  long timeInMicroseconds;

  init_vectors ();

  gettimeofday(&t1, NULL);

  for(int iter=0; iter < ITERATIONS; iter++)
    compute ();

  gettimeofday(&t2, NULL);

  timeInMicroseconds = TIME_DIFF(t1, t2);
  fprintf(stderr, "time = %g us\n", (double)timeInMicroseconds);
   
  check_result ();

  if (argc > 1)
    show_first_results ();
    
  return EXIT_SUCCESS;
}

static void check_result (void)
{
  unsigned errors = 0;

  // Compute reference vector
  for(int iter=0; iter < ITERATIONS; iter++)
    for (int i=0; i<N; i++)
      ref[i] = f(v1[i], v2[i], ref[i]);

  for (int i=0; i<N; i++) {
    if (fabs(v3[i] - ref[i]) > epsilon) {
      errors++;
      if (errors <= 10)
	printf ("%lf (computed) != %lf (expected)\n", v3[i], ref[i]);
    }
  }
  if (errors)
    fprintf (stderr, "WARNING! %d errors!\n", errors);
}

static void show_first_results (void)
{
  for (int i=0; i<10; i++)
    printf ("%f ", v3[i]);
    
  printf ("\n");
}
